@@ -17,10 +17,10 @@ Package dependencies:
 from ramp._version import __version__
 from ramp.core.core import UseCase,User,Appliance
 from ramp.core.stochastic_process import calc_peak_time_range
-from ramp.core.initialise import yearly_pattern
+from ramp.core.utils import yearly_pattern
 from ramp.example.examples import load_data,download_example
 
 __authors__ = "Listed in AUTHORS"
-__copyright__ = ".."
+__copyright__ = "Licensed under the European Union Public Licence (EUPL), Version 1.2-or-later"
 
 
